

from twisted.internet import task

from kivy_garden.ebs.core.labels import ColorLabel
from kivy_garden.ebs.core.colors import color_set_alpha

from .basemixin import BaseGuiMixin
from ebs.linuxnode.core.config import ElementSpec, ItemSpec


class NodeIDGuiMixin(BaseGuiMixin):
    _gui_nodeid_bgcolor = None
    _gui_nodeid_color = None

    def __init__(self, *args, **kwargs):
        super(NodeIDGuiMixin, self).__init__(*args, **kwargs)
        self._gui_id_tag = None
        self._gui_id_task = None

    def install(self):
        super(NodeIDGuiMixin, self).install()
        _elements = {
            'node_id_display': ElementSpec('id', 'display', ItemSpec(bool, fallback=False)),
            'node_id_display_frequency': ElementSpec('id', 'display_frequency', ItemSpec(int, fallback=0)),
            'node_id_display_duration': ElementSpec('id', 'display_duration', ItemSpec(int, fallback=15)),
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

    @property
    def gui_id_tag(self):
        if not self._gui_id_tag:
            params = {'bgcolor': (self._gui_nodeid_bgcolor or
                                  color_set_alpha(self.gui_color_1, 0.4)),
                      'color': [1, 1, 1, 1]}
            self._gui_id_tag = ColorLabel(
                text=self.id, size_hint=(None, None),
                width=250, height=50, font_size='18sp',
                valign='middle', halign='center', **params
            )
        return self._gui_id_tag

    def gui_id_show(self, duration=None):
        if not self.gui_id_tag.parent:
            self.gui_status_stack.add_widget(self.gui_id_tag)
        if duration:
            self.reactor.callLater(duration, self.gui_id_hide)

    def gui_id_hide(self):
        if self.gui_id_tag.parent:
            self.gui_status_stack.remove_widget(self.gui_id_tag)

    def gui_setup(self):
        super(NodeIDGuiMixin, self).gui_setup()
        if not self.config.node_id_display:
            return
        if self.config.node_id_display_frequency:
            self._gui_id_task = task.LoopingCall(
                self.gui_id_show,
                duration=self.config.node_id_display_duration
            )
            self._gui_id_task.start(self.config.node_id_display_frequency)
        else:
            self.gui_id_show(duration=self.config.node_id_display_duration)





from kivy.app import App
from twisted.internet import reactor

from ebs.linuxnode.gui.kivy.core.basenode import BaseIoTNodeGui


class BaseIOTNodeApplication(App):
    _node_class = BaseIoTNodeGui

    def __init__(self, config, *args, **kwargs):
        self._config = config
        self._debug = kwargs.pop('debug', False)
        super(BaseIOTNodeApplication, self).__init__(*args, **kwargs)
        self._node = None

    def build(self):
        print("Constructing Node : {}".format(self._node_class))
        self._node = self._node_class(reactor=reactor, application=self)
        print("Installing Node Resources")
        self._node.install()

        # Config is ready by this point, as long as config elements and
        # application roots are all registered in the install()
        # call-chain.
        self._config.print()

        print("Using Application Roots :")
        for root in self._config.roots:
            print("  ", root)

        print("Building GUI for node {0}".format(self._node))
        return self._node.gui_setup()

    def on_start(self):
        print("Starting Application : {}".format(self))
        self._node.start()

    def on_stop(self):
        self._node.stop()

    def on_pause(self):
        pass





from kivy_garden.ebs.progressspinner import TextureProgressSpinner
from kivy_garden.ebs.core.colors import Gradient
from .basemixin import BaseGuiMixin


class BusySpinnerGuiMixin(BaseGuiMixin):
    _gui_busy_spinner_class = TextureProgressSpinner
    _gui_busy_spinner_props = {}

    def __init__(self, *args, **kwargs):
        self._gui_busy_spinner = None
        super(BusySpinnerGuiMixin, self).__init__(*args, **kwargs)

    def busy_set(self):
        super(BusySpinnerGuiMixin, self).busy_set()
        self._gui_update_busy()

    def busy_clear(self):
        super(BusySpinnerGuiMixin, self).busy_clear()
        self._gui_update_busy()

    def _gui_update_busy(self):
        if self.busy is True:
            self._gui_busy_show()
        else:
            self._gui_busy_clear()

    def _gui_busy_show(self):
        parent = self.gui_busy_spinner.parent
        if not parent:
            self.gui_notification_row.add_widget(self.gui_busy_spinner)
            self.gui_notification_update()

    def _gui_busy_clear(self):
        parent = self.gui_busy_spinner.parent
        if parent:
            parent.remove_widget(self.gui_busy_spinner)
            self.gui_notification_update()

    @property
    def gui_busy_spinner(self):
        if not self._gui_busy_spinner:
            props = self._gui_busy_spinner_props
            _texture = Gradient.horizontal(self.gui_color_1, self.gui_color_2)
            props['texture'] = _texture
            self._gui_busy_spinner = self._gui_busy_spinner_class(
                size_hint=(None, None), height=50, pos_hint={'left': 1},
                **props
            )
        return self._gui_busy_spinner





from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.stacklayout import StackLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout
from kivy.core.window import Window

from ebs.linuxnode.core.basenode import BaseIoTNode
from ebs.linuxnode.core.config import ElementSpec
from ebs.linuxnode.core.config import ItemSpec


class BaseGuiStructureMixin(BaseIoTNode):
    def __init__(self, *args, **kwargs):
        self._gui_root = None
        self._gui_structure_root = None
        self._gui_primary_root = None
        self._gui_footer = None
        self._gui_anchor_br = None
        self._gui_anchor_bl = None
        self._gui_anchor_tr = None
        self._gui_status_stack = None
        self._gui_notification_stack = None
        self._gui_notification_row = None
        self._gui_debug_stack = None
        self._gui_content_root = None
        self._gui_main_content = None
        self._gui_sidebar = None
        self._gui_sidebar_users = set()
        self._gui_animation_layer = None
        super(BaseGuiStructureMixin, self).__init__(*args, **kwargs)

    def install(self):
        super(BaseGuiStructureMixin, self).install()
        _elements = {
            'sidebar_width': ElementSpec('display', 'sidebar_width', ItemSpec(float, fallback=0.3)),
            'sidebar_height_specific': ElementSpec('display', 'sidebar_height', ItemSpec(float, fallback=0.0)),
            'sidebar_height': ElementSpec('_derived', self._sidebar_height),
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

    def _sidebar_height(self, config):
        if config.sidebar_height_specific:
            return config.sidebar_height_specific
        else:
            return config.sidebar_width

    @property
    def gui_anchor_bottom_right(self):
        if not self._gui_anchor_br:
            self._gui_anchor_br = AnchorLayout(anchor_x='right',
                                               anchor_y='bottom',
                                               pos_hint={'pos': [0, 0]})
            self.gui_primary_root.add_widget(self._gui_anchor_br)
        return self._gui_anchor_br

    @property
    def gui_anchor_bottom_left(self):
        if not self._gui_anchor_bl:
            self._gui_anchor_bl = AnchorLayout(anchor_x='left',
                                               anchor_y='bottom',
                                               pos_hint={'pos': [0, 0]})
            self.gui_primary_root.add_widget(self._gui_anchor_bl)
        return self._gui_anchor_bl

    @property
    def gui_anchor_top_right(self):
        if not self._gui_anchor_tr:
            self._gui_anchor_tr = AnchorLayout(anchor_x='right',
                                               anchor_y='top',
                                               pos_hint={'pos': [0, 0]})
            self.gui_primary_root.add_widget(self._gui_anchor_tr)
        return self._gui_anchor_tr

    @property
    def gui_status_stack(self):
        if not self._gui_status_stack:
            self._gui_status_stack = StackLayout(orientation='bt-rl',
                                                 padding='8sp')
            self.gui_anchor_bottom_right.add_widget(self._gui_status_stack)
        return self._gui_status_stack

    @property
    def gui_notification_stack(self):
        if not self._gui_notification_stack:
            self._gui_notification_stack = GridLayout(cols=1,
                                                      padding='8sp',
                                                      spacing='8sp',
                                                      size_hint_y=None,)

            def _set_height(_, mheight):
                self.gui_notification_stack.height = mheight
            self.gui_notification_stack.bind(minimum_height=_set_height)
            self.gui_anchor_bottom_left.add_widget(self._gui_notification_stack)
        return self._gui_notification_stack

    @property
    def gui_notification_row(self):
        if not self._gui_notification_row:
            self._gui_notification_row = StackLayout(orientation='lr-bt',
                                                     spacing='8sp')
            self.gui_notification_stack.add_widget(self._gui_notification_row)
        return self._gui_notification_row

    def gui_notification_update(self):
        self.gui_notification_row.do_layout()
        self.gui_notification_stack.do_layout()

    @property
    def gui_debug_stack(self):
        if not self._gui_debug_stack:
            self._gui_debug_stack = StackLayout(orientation='tb-rl',
                                                padding='8sp')
            self.gui_anchor_top_right.add_widget(self._gui_debug_stack)
        return self._gui_debug_stack

    @property
    def gui_footer(self):
        if not self._gui_footer:
            _ = self.gui_primary_root
            self._gui_footer = BoxLayout(
                orientation='vertical', size_hint=(1, None),
                height=80, padding=['0sp', '0sp', '0sp', '8sp']
            )
        return self._gui_footer

    def gui_footer_show(self):
        if not self._gui_footer.parent:
            self.gui_structure_root.add_widget(self._gui_footer)

    def gui_footer_hide(self):
        if self._gui_footer.parent:
            self.gui_structure_root.remove_widget(self._gui_footer)

    @property
    def gui_primary_root(self):
        if not self._gui_primary_root:
            self._gui_primary_root = FloatLayout()
            self.gui_structure_root.add_widget(self._gui_primary_root)
        return self._gui_primary_root

    @property
    def gui_structure_root(self):
        if not self._gui_structure_root:
            self._gui_structure_root = BoxLayout(orientation='vertical')
            self.gui_root.add_widget(self._gui_structure_root)
        return self._gui_structure_root

    @property
    def gui_main_content(self):
        if not self._gui_main_content:
            self._gui_main_content = RelativeLayout()
            self.gui_content_root.add_widget(self._gui_main_content)
        return self._gui_main_content

    @property
    def gui_sidebar(self):
        if not self._gui_sidebar:
            if self.config.portrait or self.config.os_rotation:
                orientation = "horizontal"
                y_hint = self.config.sidebar_height / (1 - self.config.sidebar_height)
                size_hint = (1, y_hint)
            else:
                orientation = "vertical"
                x_hint = self.config.sidebar_width / (1 - self.config.sidebar_width)
                size_hint = (x_hint, 1)
            self._gui_sidebar = BoxLayout(orientation=orientation,
                                          size_hint=size_hint)
        return self._gui_sidebar

    def gui_sidebar_show(self, key):
        if not key:
            key = 'unspecified'
        self._gui_sidebar_users.add(key)
        if not self.gui_sidebar.parent:
            self.log.debug("Showing sidebar")
            self.gui_content_root.add_widget(self.gui_sidebar)

    def gui_sidebar_hide(self, key):
        if key not in self._gui_sidebar_users:
            return
        self._gui_sidebar_users.remove(key)
        if len(self._gui_sidebar_users):
            return
        if self.gui_sidebar.parent:
            self.log.debug("Hiding sidebar")
            self.gui_content_root.remove_widget(self.gui_sidebar)

    @property
    def gui_content_root(self):
        if not self._gui_content_root:
            params = {'spacing': 0}
            if self.config.portrait or self.config.os_rotation:
                params['orientation'] = 'vertical'
            else:
                params['orientation'] = 'horizontal'
            self._gui_content_root = BoxLayout(**params)
            self.gui_root.add_widget(self._gui_content_root,
                                     len(self.gui_root.children) - 1)
        return self._gui_content_root

    @property
    def gui_animation_layer(self):
        if not self._gui_animation_layer:
            self._gui_animation_layer = FloatLayout()
            self.gui_root.add_widget(self._gui_animation_layer,
                                     len(self.gui_root.children) - 1)
        return self._gui_animation_layer

    @property
    def gui_root(self):
        if not self._gui_root:
            self._gui_root = FloatLayout()
        return self._gui_root

    @property
    def window_size(self):
        return Window.size

    @property
    def window_height(self):
        return self.window_size[1]

    @property
    def window_width(self):
        return self.window_size[0]

    def geometry_transform(self, x, y, width, height):
        if self.config.os_rotation:
            return x, y, width, height
        if self.config.flip:
            x = self.window_width - width - x
            y = self.window_height - height - y
        if self.config.portrait:
            x, y, width, height = y, x, height, width
        return x, y, width, height




from twisted.logger import formatEvent

from datetime import datetime
from collections import deque
from kivy.core.window import Window
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.uix.scrollview import ScrollView
from kivy.effects.scroll import ScrollEffect
from kivy.utils import get_hex_from_color

from kivy_garden.ebs.core.labels import ColorLabel
from ebs.linuxnode.core.log import NodeLoggingMixin
from ebs.linuxnode.core.config import ElementSpec, ItemSpec
from .basemixin import BaseGuiMixin


class LoggingGuiMixin(BaseGuiMixin):
    def __init__(self, *args, **kwargs):
        self._gui_log = None
        self._gui_log_end = None
        self._gui_log_layout = None
        self._gui_log_scroll = None
        self._gui_log_lines = deque([], maxlen=100)
        super(LoggingGuiMixin, self).__init__(*args, **kwargs)

    def install(self):
        super(LoggingGuiMixin, self).install()
        _elements = {
            'gui_log_display': ElementSpec('debug', 'gui_log_display', ItemSpec(bool, fallback=False)),
            'gui_log_level': ElementSpec('debug', 'gui_log_level', ItemSpec(fallback='info')),
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

    def _observers(self):
        rv = NodeLoggingMixin._observers(self)
        if self.config.gui_log_display:
            rv.extend([self.gui_log_observer])
        return rv

    _level_ignore_map = {
        'trace': [],
        'debug': ['trace'],
        'info': ['trace', 'debug'],
        'warning': ['trace', 'debug', 'info'],
        'error': ['trace', 'debug', 'info', 'error'],
    }

    def gui_log_observer(self, event):
        ll = event['log_level'].name
        if ll in self._level_ignore_map[self.config.gui_log_level]:
            return
        msg = "[font=RobotoMono-Regular][{0:^8}][/font] {1} {2}".format(
            ll.upper(),
            datetime.fromtimestamp(event['log_time']).strftime("%d%m %H:%M:%S"),
            formatEvent(event)
        )
        color = None
        if ll == 'warn':
            color = [1, 1, 0, 1]
        elif ll == 'error':
            color = [1, 0, 0, 1]
        elif ll == 'critical':
            color = [1, 0, 0, 1]
        if color:
            color = get_hex_from_color(color)
            msg = '[color={0}]{1}[/color]'.format(color, msg)

        self._gui_log_lines.append(msg)
        self._gui_log.text = '\n'.join(self._gui_log_lines)

    @property
    def gui_log(self):
        if not self._gui_log:
            self._gui_log_scroll = ScrollView(
                size_hint=(None, None), bar_pos_y='right',
                bar_color=[1, 1, 1, 1], effect_cls=ScrollEffect,
                do_scroll_x=False, do_scroll_y=True)

            def _set_log_size(_, size):
                width = min(max(700, size[0] * 0.3), size[0])
                height = size[1] * 0.6
                self._gui_log_scroll.size = width, height
            self.gui_root.bind(size=_set_log_size)

            self._gui_log_layout = BoxLayout(orientation='vertical',
                                             size_hint=(1, None))

            self._gui_log = ColorLabel(
                size_hint=(1, 1), padding=(8, 8), bgcolor=[0, 0, 0, 0.2],
                halign='left', valign='top', markup=True, font_size='12sp',
            )

            self._gui_log_end = Label(size_hint=(1, None), height=1)

            def _set_label_height(_, size):
                if size[1] > Window.height * 0.3:
                    self._gui_log.height = size[1]
                else:
                    self._gui_log.height = Window.height * 0.3
            self._gui_log.bind(texture_size=_set_label_height)

            def _set_layout_height(_, size):
                if size[1] > Window.height * 0.3:
                    self._gui_log_layout.height = size[1]
                else:
                    self._gui_log_layout.height = Window.height * 0.3
                self._gui_log_scroll.scroll_to(self._gui_log_end)
            self._gui_log.bind(texture_size=_set_layout_height)

            def _set_text_width(_, size):
                self._gui_log.text_size = size[0], None
            self._gui_log.bind(size=_set_text_width)

            self._gui_log_layout.add_widget(self._gui_log)
            self._gui_log_layout.add_widget(self._gui_log_end)
            self._gui_log_scroll.add_widget(self._gui_log_layout)
            self.gui_debug_stack.add_widget(self._gui_log_scroll)
        return self._gui_log

    def gui_setup(self):
        super(LoggingGuiMixin, self).gui_setup()
        if not self.config.gui_log_display:
            return
        _ = self.gui_log





import os
import subprocess
from kivy.core.window import Window
from ebs.linuxnode.core.config import ElementSpec, ItemSpec

from .sequence import BackgroundSequenceMixin


class OverlayWindowGuiMixin(BackgroundSequenceMixin):
    # Overlay mode needs specific host support.
    # At present, this never works for all practical purposes.
    # RPi :
    #   See DISPMANX layers and
    #   http://codedesigner.de/articles/omxplayer-kivy-overlay/index.html
    # Normal Linux Host :
    #   See core-x11 branch and
    #   https://groups.google.com/forum/#!topic/kivy-users/R4aJCph_7IQ
    # Others :
    #   Unknown, see
    #   - https://github.com/kivy/kivy/issues/4307
    #   - https://github.com/kivy/kivy/pull/5252
    _gui_supports_overlay_mode = True

    def __init__(self, *args, **kwargs):
        self._overlay_mode = None
        self._foundation_process = None
        super(OverlayWindowGuiMixin, self).__init__(*args, **kwargs)

    def install(self):
        super(OverlayWindowGuiMixin, self).install()
        _elements = {
            'overlay_mode': ElementSpec('display', 'overlay_mode', ItemSpec(bool, fallback=False)),
            'show_foundation': ElementSpec('display-rpi', 'show_foundation', ItemSpec(bool, fallback=False)),
            'dispmanx_foundation_layer': ElementSpec('display-rpi', 'dispmanx_foundation_layer',
                                                     ItemSpec(int, fallback=-250)),
            'foundation_image': ElementSpec('display-rpi', 'foundation_image', ItemSpec(fallback=None)),
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

    @property
    def overlay_mode(self):
        return self._overlay_mode

    @overlay_mode.setter
    def overlay_mode(self, value):
        if not self._gui_supports_overlay_mode and value:
            self.log.warn("Application tried to change overlay mode, "
                          "not supported this platform.")
            return
        if value is True:
            self._gui_overlay_mode_enter()
        else:
            self._gui_overlay_mode_exit()

    def _gui_overlay_mode_enter(self):
        self.log.info('Entering Overlay Mode')
        if self._overlay_mode:
            return
        self._overlay_mode = True
        self.gui_bg_pause()
        Window.clearcolor = [0, 0, 0, 0]

    def _gui_overlay_mode_exit(self):
        self.log.info('Exiting Overlay Mode')
        if not self._overlay_mode:
            return
        self._overlay_mode = False
        Window.clearcolor = [0, 0, 0, 1]
        self.gui_bg_resume()

    def stop(self):
        if self._foundation_process:
            self._foundation_process.terminate()
        super(OverlayWindowGuiMixin, self).stop()

    def gui_setup(self):
        super(OverlayWindowGuiMixin, self).gui_setup()

        if self.config.show_foundation and \
                self.config.foundation_image and \
                os.path.exists(self.config.foundation_image):
            cmd = ['pngview', '-l', '"{}"'.format(str(self.config.dispmanx_foundation_layer)),
                   '-n', self.config.foundation_image]
            self._foundation_process = subprocess.Popen(cmd)

        self.overlay_mode = self.config.overlay_mode





from ebs.linuxnode.core.config import ElementSpec, ItemSpec
from .basemixin import BaseGuiMixin


class DisplayMixin(BaseGuiMixin):
    def install(self):
        super(DisplayMixin, self).install()
        _elements = {
            'fullscreen': ElementSpec('display', 'fullscreen', ItemSpec(bool, fallback=True)),
            'portrait': ElementSpec('display', 'portrait', ItemSpec(bool, read_only=False, fallback=False)),
            'flip': ElementSpec('display', 'flip', ItemSpec(bool, read_only=False, fallback=False)),
            'orientation': ElementSpec('_derived', self._orientation),
            'os_rotation': ElementSpec('display', 'os_rotation', ItemSpec(bool, fallback=False)),
            'app_dispmanx_layer': ElementSpec('display-rpi', 'dispmanx_app_layer', ItemSpec(int, fallback=5))
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

    def _orientation(self, config):
        rv = 0
        if config.portrait is True:
            rv += 90
        if config.flip:
            rv += 180
        return rv

    def orientation_update(self):
        from kivy.config import Config
        Config.set('graphics', 'rotation', self.config.orientation)





import os
from ebs.linuxnode.gui.kivy.background.mixin import OverlayWindowGuiMixin
from .text import FontsGuiMixin
from .log import LoggingGuiMixin
from .busy import BusySpinnerGuiMixin
from .nodeid import NodeIDGuiMixin
from .display import DisplayMixin


class BaseIoTNodeGui(NodeIDGuiMixin,
                     BusySpinnerGuiMixin,
                     LoggingGuiMixin,
                     FontsGuiMixin,
                     OverlayWindowGuiMixin,
                     DisplayMixin):

    def __init__(self, *args, **kwargs):
        self._application = kwargs.pop('application')
        self._gui_root = None
        super(BaseIoTNodeGui, self).__init__(*args, **kwargs)

    def install(self):
        super(BaseIoTNodeGui, self).install()
        self.config.register_application_root(
            os.path.abspath(os.path.dirname(__file__))
        )

    def sysinfo_install(self):
        super(BaseIoTNodeGui, self).sysinfo_install()
        self.sysinfo.app.versions.register_package('kivy')
        self.sysinfo.app.versions.register_namespace('kivy-garden')


def gui_setup(self):
        super(BaseIoTNodeGui, self).gui_setup()
        return self.gui_root





import os
from ebs.linuxnode.gui.kivy.utils.application import BaseIOTNodeApplication
from node import ExampleNode


class ExampleApplication(BaseIOTNodeApplication):
    _node_class = ExampleNode

    def build(self):
        # This is an emergency-only approach. In general, configure
        # roots in the appropriate node classes instead.
        # ( for ex, see BaseIoTNodeGui.install() )
        r = super(ExampleApplication, self).build()
        self._config.register_application_root(
            os.path.abspath(os.path.dirname(__file__))
        )
        return r

    def on_start(self):
        # Config is ready by this point, as long as config elements and
        # application roots are all registered in the install()
        # call-chain.
        self._config.print()
        # Application Roots should also be ready before this
        super(ExampleApplication, self).on_start()





from .base import BackgroundProviderBase


class StructuredBackgroundProvider(BackgroundProviderBase):
    def check_support(self, target):
        if not isinstance(target, str):
            return False
        if not target.startswith('structured:'):
            return False
        if not hasattr(self.actual, target.split(':')[1]):
            return False
        return True

    def play(self, target, duration=None, callback=None, **kwargs):
        _target = getattr(self.actual, target.split(':')[1])
        if callable(_target):
            self._widget = _target(**kwargs)
        else:
            self._widget = _target
        super(StructuredBackgroundProvider, self).play(target, duration, callback, **kwargs)
        return self._widget

    def stop(self):
        if hasattr(self._widget, 'stop'):
            self._widget.stop()
        super(StructuredBackgroundProvider, self).stop()

    def pause(self):
        super(StructuredBackgroundProvider, self).pause()
        if hasattr(self._widget, 'pause'):
            self._widget.pause()

    def resume(self):
        if hasattr(self._widget, 'retrigger'):
            self._widget.retrigger()
        super(StructuredBackgroundProvider, self).resume()





import os
from raspi_system import hwinfo

from ebs.linuxnode.core.config import ItemSpec
from ebs.linuxnode.core.config import ElementSpec


def _orientation(config):
    rv = 0
    if config.portrait is True:
        rv += 90
    if config.flip:
        rv += 180
    return rv


def prepare_config(appname):
    from ebs.linuxnode.core.config import IoTNodeConfig
    node_config = IoTNodeConfig(appname=appname)

    items = [
            ('platform', ElementSpec('platform', 'platform', ItemSpec(fallback='native'))),
            ('fullscreen', ElementSpec('display', 'fullscreen', ItemSpec(bool, fallback=True))),
            ('portrait', ElementSpec('display', 'portrait', ItemSpec(bool, fallback=False))),
            ('flip', ElementSpec('display', 'flip', ItemSpec(bool, fallback=False))),
            ('app_dispmanx_layer', ElementSpec('display-rpi', 'dispmanx_app_layer', ItemSpec(int, fallback=5))),
            ('orientation', ElementSpec('_derived', _orientation)),
            ('os_rotation', ElementSpec('display', 'os_rotation', ItemSpec(bool, fallback=False)))
    ]

    for name, spec in items:
        node_config.register_element(name, spec)
    return node_config


def prepare_environment(node_config):
    print("Using pango Text Provider")
    os.environ['KIVY_TEXT'] = 'pango'
    if node_config.platform == 'rpi':
        if hwinfo.is_pi4():
            print("Using sdl2 Window Provider")
            os.environ['KIVY_WINDOW'] = 'sdl2'
        else:
            print("Using egl_rpi Window Provider")
            os.environ['KIVY_WINDOW'] = 'egl_rpi'
        os.environ['KIVY_BCM_DISPMANX_LAYER'] = str(node_config.app_dispmanx_layer)
        print("Using app_dispmanx_layer {0}".format(node_config.app_dispmanx_layer))
    else:
        print("Using ffpyplayer Video Provider")
        os.environ['KIVY_VIDEO'] = 'ffpyplayer'


def prepare_kivy(node_config):
    from kivy.config import Config
    if node_config.fullscreen is True:
        Config.set('graphics', 'fullscreen', 'auto')

    if node_config.orientation:
        Config.set('graphics', 'rotation', node_config.orientation)

    Config.set('kivy', 'keyboard_mode', 'systemandmulti')
    Config.set('input', 'mouse', 'mouse,multitouch_on_demand')

    from kivy.support import install_twisted_reactor
    install_twisted_reactor()





import time


class BackgroundProviderBase(object):
    is_visual = True

    def __init__(self, actual):
        self._actual = actual
        self._widget = None
        self._end_call = None
        self._paused = False
        self._eresidual = None
        self._callback = None

    @property
    def actual(self):
        if hasattr(self._actual, 'actual'):
            return self._actual.actual
        else:
            return self._actual

    def check_support(self, target):
        # Check if the provider supports the target and
        # if the target exists.
        raise NotImplementedError

    def play(self, target, duration=None, callback=None, **kwargs):
        # Create a Widgetized Background and return it.
        # It will be attached later.
        if duration and callback:
            self._callback = callback
            self._end_call = self.actual.reactor.callLater(duration, callback)

    def stop(self):
        # Stop and unload the Widgetized Background.
        # The widget has already been detached.
        if self._end_call and self._end_call.active():
            self._end_call.cancel()
        self._widget = None

    def pause(self):
        # Pause the Widgetized Background.
        # It has already been detached.
        if self._paused:
            return
        if self._end_call and self._end_call.active():
            ietime = self._end_call.getTime()
            ptime = time.time()
            self._eresidual = ietime - ptime
            self._end_call.cancel()
        self._paused = True

    def resume(self):
        # Resume the Widgetized Background.
        # It will be attached later.
        if not self._paused:
            return
        self._paused = False
        if self._eresidual:
            self._end_call = self.actual.reactor.callLater(self._eresidual, self._callback)
            self._eresidual = None





from numbers import Number
from kivy_garden.ebs.core.colors import ColorBoxLayout

from .base import BackgroundProviderBase


class ColorBackgroundProvider(BackgroundProviderBase):
    def _parse_color_str(self, target):
        color = target.split(':')
        if len(color) not in (3, 4):
            raise ValueError
        try:
            color = (float(x) for x in color)
        except ValueError:
            raise
        return color

    def check_support(self, target):
        if isinstance(target, str):
            try:
                self._parse_color_str(target)
                return True
            except ValueError:
                return False
        elif isinstance(target, (tuple, list)):
            if len(target) not in (3, 4):
                return False
            for elem in target:
                if not isinstance(elem, Number):
                    return False
        else:
            return False
        return True

    def play(self, target, duration=None, callback=None, **kwargs):
        if isinstance(target, str):
            target = self._parse_color_str(target)
        self._widget = ColorBoxLayout(bgcolor=target)
        super(ColorBackgroundProvider, self).play(target, duration, callback, **kwargs)
        return self._widget





import os
from kivy.core.text import FontContextManager
from ebs.linuxnode.core.config import ElementSpec, ItemSpec
from .basemixin import BaseGuiMixin


class FontsGuiMixin(BaseGuiMixin):
    def __init__(self, *args, **kwargs):
        self._text_font_context = None
        super(FontsGuiMixin, self).__init__(*args, **kwargs)

    def install(self):
        super(FontsGuiMixin, self).install()
        _elements = {
            'text_font_name': ElementSpec('text', 'font_name', ItemSpec('path', fallback=None)),
            'text_use_fcm': ElementSpec('text', 'use_fcm', ItemSpec(bool, fallback=False)),
            'text_fcm_system': ElementSpec('text', 'fcm_system', ItemSpec(bool, fallback=True)),
            'text_fcm_fonts': ElementSpec('text', 'fcm_fonts', ItemSpec('path', fallback=None)),
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

    @property
    def text_font_context(self):
        if not self._text_font_context and self.config.text_use_fcm:
            self._text_create_fcm()
        return self._text_font_context

    def _text_create_fcm(self):
        fc = self.appname
        if self.config.text_fcm_system:
            fc = "system://{0}".format(fc)
        self._text_font_context = fc
        self.log.info("Creating FontContextManager {0} using fonts in {1}"
                      .format(fc, self.config.text_fcm_fonts))
        FontContextManager.create(fc)

        for filename in os.listdir(self.config.text_fcm_fonts):
            self.log.info("Installing Font {0} to FCM {1}".format(filename, self._text_font_context))
            FontContextManager.add_font(fc, os.path.join(self.config.text_fcm_fonts, filename))

    @property
    def text_font_params(self):
        params = {}
        if self.text_font_context:
            params.update({
                'font_context': self._text_font_context
            })
        else:
            params.update({
                'font_name': self.config.text_font_name
            })
        return params





import os
import shutil
import appdirs
import warnings
from collections import namedtuple

from kivy.uix.boxlayout import BoxLayout

from ebs.linuxnode.core.config import ElementSpec, ItemSpec
from ebs.linuxnode.gui.kivy.core.basemixin import BaseGuiMixin

from .base import BackgroundProviderBase
from .image import ImageBackgroundProvider
from .color import ColorBackgroundProvider
from .structured import StructuredBackgroundProvider


# TODO This is here, but not particularly tested. Expect it to only work for video.
# In general, background sequencing should be handled outside the Background code or
# by dedicated infrastructure with effectively allows gui_bg to accept a list of this class.
BackgroundSpec = namedtuple('BackgroundSpec', ["target", "bgcolor", "callback", "duration"],
                            defaults=[None, None, None])


class BackgroundGuiMixin(BaseGuiMixin):
    def __init__(self, *args, **kwargs):
        self._bg_providers = []
        self._bg_container = None
        self._bg = None
        self._bg_current = None
        self._bg_current_provider = None
        super(BackgroundGuiMixin, self).__init__(*args, **kwargs)

    def install_background_provider(self, provider):
        self.log.info("Installing BG Provider {}".format(provider))
        self._bg_providers.insert(0, provider)

    def install(self):
        super(BackgroundGuiMixin, self).install()

        _path = os.path.abspath(os.path.dirname(__file__))
        fallback_default = os.path.join(_path, 'images/background.png')
        fallback = os.path.join(appdirs.user_config_dir(self.config.appname), 'background.png')
        if not os.path.exists(fallback):
            shutil.copy(fallback_default, fallback)

        _elements = {
            'image_bgcolor': ElementSpec('display', 'image_bgcolor', ItemSpec('kivy_color', fallback='auto')),
            'background': ElementSpec('display', 'background', ItemSpec(str, read_only=False, fallback=fallback)),
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

        self.install_background_provider(ColorBackgroundProvider(self))
        self.install_background_provider(ImageBackgroundProvider(self))
        self.install_background_provider(StructuredBackgroundProvider(self))

    def _get_provider(self, target):
        provider: BackgroundProviderBase
        provider = None
        for lprovider in self._bg_providers:
            if lprovider.check_support(target):
                provider = lprovider
                break
        return provider

    def background_set(self, target):
        if not target:
            target = None

        provider = self._get_provider(target)
        if not provider:
            self.log.warn("Provider not found for background {}. Not Setting.".format(target))
            target = None

        if target and self.config.background != target:
            self.config.background = target

        self.gui_bg_update()

    @property
    def gui_bg_container(self):
        if self._bg_container is None:
            self._bg_container = BoxLayout()
            self.gui_main_content.add_widget(self._bg_container)
        return self._bg_container

    def gui_bg_clear(self):
        if self._bg and self._bg.parent:
            self.gui_bg_container.remove_widget(self._bg)
        self._bg = None
        if self._bg_current_provider:
            self._bg_current_provider.stop()

    @property
    def gui_bg(self):
        return self._bg_current

    @gui_bg.setter
    def gui_bg(self, value):
        self.log.info("Setting background to {value}", value=value)
        bgcolor, callback, duration = None, None, None
        if isinstance(value, BackgroundSpec):
            value, bgcolor, callback, duration = value

        if not bgcolor:
            bgcolor = self.config.image_bgcolor

        provider = self._get_provider(value)

        if not provider:
            self.log.warn("Provider not found for background {}".format(value))
            value = self.config.background
            provider = self._get_provider(value)
            self.log.warn("Using {} instead.".format(value))

        self.log.debug("Using {} to show background {}".format(provider, value))
        self.gui_bg_clear()

        self._bg_current = value
        self._bg_current_provider = provider
        self._bg = self._bg_current_provider.play(
            value, bgcolor=bgcolor, callback=callback, duration=duration
        )
        self.gui_bg_container.add_widget(self._bg)

    def gui_bg_pause(self):
        self.log.debug("Pausing Background")
        self.gui_main_content.remove_widget(self._bg_container)
        if self._bg_current_provider:
            self._bg_current_provider.pause()

    def gui_bg_resume(self):
        self.log.debug("Resuming Background")
        if self._bg_current_provider:
            self._bg_current_provider.resume()
        if not self._bg_container.parent:
            self.gui_main_content.add_widget(self._bg_container, len(self.gui_main_content.children))

    def stop(self):
        if self._bg_current_provider:
            self._bg_current_provider.stop()
        super(BackgroundGuiMixin, self).stop()

    def gui_bg_update(self):
        self.gui_bg = self.config.background

    def gui_setup(self):
        super(BackgroundGuiMixin, self).gui_setup()
        self.gui_bg_update()





from itertools import cycle

from .manager import BackgroundGuiMixin
from .manager import BackgroundSpec


class BackgroundSequenceMixin(BackgroundGuiMixin):
    def __init__(self, *args, **kwargs):
        super(BackgroundSequenceMixin, self).__init__(*args, **kwargs)
        self._bg_sequence = cycle([])

    @property
    def gui_bg_sequence(self):
        return self._bg_sequence

    @gui_bg_sequence.setter
    def gui_bg_sequence(self, value):
        self._bg_sequence = cycle(value)
        self.gui_bg_step()

    def gui_bg_step(self, *_):
        target = next(self.gui_bg_sequence)
        bgcolor, callback, duration = None, None, None
        if isinstance(target, BackgroundSpec):
            target, bgcolor, callback, duration = target

        if not bgcolor:
            bgcolor = self.config.image_bgcolor

        if callback:
            self.log.warn("BG Sequence recieved an item with a callback. "
                          "This is not supported and is ignored. {}".format(target))

        callback = self.gui_bg_step
        spec = BackgroundSpec(target, bgcolor, callback, duration)
        self.log.debug("BG Sequence Step : {}".format(spec))
        self.gui_bg = spec

    def background_sequence_set(self, targets):
        if not targets:
            targets = []

        for target in targets:
            provider = self._get_provider(target)
            if not provider:
                self.log.warn("Provider not found for background {}. Not Setting.".format(target))
                targets.remove(target)

        # TODO Establish Peristence

        self.gui_bg_update()

    def gui_bg_update(self):
        if False:
            self.gui_bg_sequence = []
        else:
            super(BackgroundSequenceMixin, self).gui_bg_update()





import os
from kivy_garden.ebs.core.image import BleedImage

from .base import BackgroundProviderBase


class ImageBackgroundProvider(BackgroundProviderBase):
    def check_support(self, target):
        if not isinstance(target, str):
            return False
        if not os.path.exists(target):
            return False
        _extentions = ('.png', '.jpg', '.bmp', '.gif', '.jpeg')
        if os.path.splitext(target)[1] not in _extentions:
            return False
        return True

    def play(self, target, duration=None, callback=None, **kwargs):
        self._widget = BleedImage(
            source=target,
            allow_stretch=True,
            keep_ratio=True,
            **kwargs
        )
        super(ImageBackgroundProvider, self).play(target, duration, callback, **kwargs)
        return self._widget



import setuptools

_requires = [
    'six',
    'setuptools-scm',
    'appdirs',
    'kivy>=2.1.0',
    'kivy-garden',

    # ebs Widgets
    'kivy_garden.ebs.core>=1.3.0',
    'kivy_garden.ebs.progressspinner>=1.2.0',

    'ebs-linuxnode-core>=2.4.0',
]

setuptools.setup(
    name='ebs-linuxnode-gui-kivy-core',
    url='https://github.com/ebs-universe/ebs-linuxnode-coregui-kivy',

    author='Chintalagiri Shashank',
    author_email='shashank.chintalagiri@gmail.com',

    description='Kivy GUI Core for EBS Linuxnode Applications',
    long_description='',

    packages=setuptools.find_packages(),
    package_dir={'ebs.linuxnode.gui.kivy.core': 'ebs/linuxnode/gui/kivy/core'},
    package_data={'ebs.linuxnode.gui.kivy.core': ['images/background.png',
                                                  'images/no-internet.png',
                                                  'images/no-server.png']},

    install_requires=_requires,

    setup_requires=['setuptools_scm'],
    use_scm_version=True,

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Operating System :: POSIX :: Linux',
    ],
)





from .structure import BaseGuiStructureMixin
from kivy_garden.ebs.core.colors import GuiPalette


class BaseGuiMixin(BaseGuiStructureMixin):
    _palette = GuiPalette(
        background=(0x00 / 255, 0x00 / 255, 0x00 / 255),
        foreground=(0xff / 255, 0xff / 255, 0xff / 255),
        color_1=(0x00 / 255, 0x00 / 255, 0xff / 255),
        color_2=(0xff / 255, 0x00 / 255, 0x00 / 255)
    )

    def gui_setup(self):
        pass

    @property
    def gui_color_1(self):
        return self._palette.color_1

    @property
    def gui_color_2(self):
        return self._palette.color_2

    @property
    def gui_color_foreground(self):
        return self._palette.foreground

    @property
    def gui_color_background(self):
        return self._palette.background





from twisted.internet import reactor
from kivy_garden.ebs.clocks.digital import SimpleDigitalClock
from ebs.linuxnode.gui.kivy.core.basenode import BaseIoTNodeGui


class ExampleNode(BaseIoTNodeGui):
    @property
    def clock(self):
        return SimpleDigitalClock()

    def _set_bg(self, target):
        self.gui_bg = target

    def _background_examples(self):
        reactor.callLater(10, self._set_bg, '1.0:0.5:0.5:1.0')
        reactor.callLater(20, self._set_bg, 'image.jpg')
        reactor.callLater(30, self._set_bg, '0.5:1.0:0.5:1.0')
        reactor.callLater(40, self._set_bg, None)
        # Install kivy_garden.ebs.clocks
        # reactor.callLater(50, self._set_bg, 'structured:clock')

    def _enter_overlay_mode(self):
        self.overlay_mode = True

    def _exit_overlay_mode(self):
        self.overlay_mode = False

    def _overlay_examples(self):
        reactor.callLater(10, self._enter_overlay_mode)
        reactor.callLater(30, self._exit_overlay_mode)

    def start(self):
        super(ExampleNode, self).start()
        self._overlay_examples()
